from functools import lru_cache

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_prefix="", extra="ignore")

    mock_port: int = Field(default=8000, validation_alias="MOCK_PORT")
    mock_default_model: str = Field(default="mock-chat-v1", validation_alias="MOCK_DEFAULT_MODEL")
    mock_chunk_delay_ms: int = Field(default=30, ge=0, validation_alias="MOCK_CHUNK_DELAY_MS")
    mock_embedding_dims: int = Field(default=1536, ge=1, validation_alias="MOCK_EMBEDDING_DIMS")
    mock_require_auth: bool = Field(default=False, validation_alias="MOCK_REQUIRE_AUTH")
    mock_chaos_rate: float = Field(
        default=0.0,
        ge=0.0,
        le=1.0,
        validation_alias="MOCK_CHAOS_RATE",
    )
    mock_log_requests: bool = Field(default=True, validation_alias="MOCK_LOG_REQUESTS")

    @property
    def chunk_delay_seconds(self) -> float:
        return self.mock_chunk_delay_ms / 1000


@lru_cache
def get_settings() -> Settings:
    return Settings()



from __future__ import annotations

import json

import pytest


def test_json_mode_uses_structured_tool_schema(client):
    response = client.post(
        "/v1/chat/completions",
        json={
            "model": "gpt-4o",
            "response_format": {"type": "json_object"},
            "messages": [{"role": "user", "content": "Return structured json"}],
            "tools": [
                {
                    "type": "function",
                    "function": {
                        "name": "make_user",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "name": {"type": "string"},
                                "age": {"type": "integer"},
                                "score": {"type": "number"},
                                "active": {"type": "boolean"},
                                "tags": {"type": "array", "items": {"type": "string"}},
                            },
                        },
                    },
                }
            ],
        },
    )

    assert response.status_code == 200
    content = response.json()["choices"][0]["message"]["content"]
    body = json.loads(content)
    assert isinstance(body["name"], str)
    assert isinstance(body["age"], int)
    assert isinstance(body["score"], float)
    assert isinstance(body["active"], bool)
    assert isinstance(body["tags"], list)


def test_json_mode_uses_response_json_schema(client):
    response = client.post(
        "/v1/chat/completions",
        json={
            "model": "gpt-4o",
            "response_format": {
                "type": "json_schema",
                "json_schema": {
                    "name": "city_schema",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "city": {"type": "string", "enum": ["Amsterdam", "Utrecht"]},
                            "count": {"type": "integer"},
                        },
                    },
                },
            },
            "messages": [{"role": "user", "content": "Return structured json"}],
        },
    )

    assert response.status_code == 200
    content = response.json()["choices"][0]["message"]["content"]
    body = json.loads(content)
    assert body["city"] == "Amsterdam"
    assert body["count"] == 42


def test_json_mode_respects_string_formats(client):
    response = client.post(
        "/v1/chat/completions",
        json={
            "model": "gpt-4o",
            "response_format": {
                "type": "json_schema",
                "json_schema": {
                    "name": "formatted_strings",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "birth_date": {"type": "string", "format": "date"},
                            "created_at": {"type": "string", "format": "date-time"},
                            "email": {"type": "string", "format": "email"},
                            "homepage": {"type": "string", "format": "uri"},
                            "avatar": {"type": "string", "format": "url"},
                            "id": {"type": "string", "format": "uuid"},
                        },
                    },
                },
            },
            "messages": [{"role": "user", "content": "Return structured json"}],
        },
    )

    assert response.status_code == 200
    content = response.json()["choices"][0]["message"]["content"]
    body = json.loads(content)
    assert body == {
        "birth_date": "2026-01-01",
        "created_at": "2026-01-01T00:00:00",
        "email": "mock@example.com",
        "homepage": "https://example.com",
        "avatar": "https://example.com",
        "id": "00000000-0000-0000-0000-000000000000",
    }


@pytest.mark.parametrize("key", ["anyOf", "oneOf"])
def test_json_mode_prefers_non_null_variant(client, key):
    response = client.post(
        "/v1/chat/completions",
        json={
            "model": "gpt-4o",
            "response_format": {
                "type": "json_schema",
                "json_schema": {
                    "name": "nullable_field",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "value": {
                                key: [
                                    {"type": "null"},
                                    {"type": "integer"},
                                ]
                            }
                        },
                    },
                },
            },
            "messages": [{"role": "user", "content": "Return structured json"}],
        },
    )

    assert response.status_code == 200
    content = response.json()["choices"][0]["message"]["content"]
    body = json.loads(content)
    assert body["value"] == 42


def test_json_mode_generates_additional_properties_values(client):
    response = client.post(
        "/v1/chat/completions",
        json={
            "model": "gpt-4o",
            "response_format": {
                "type": "json_schema",
                "json_schema": {
                    "name": "dict_of_arrays",
                    "schema": {
                        "type": "object",
                        "additionalProperties": {
                            "type": "array",
                            "items": {"type": "integer"},
                        },
                    },
                },
            },
            "messages": [{"role": "user", "content": "Return structured json"}],
        },
    )

    assert response.status_code == 200
    content = response.json()["choices"][0]["message"]["content"]
    body = json.loads(content)
    assert body == {
        "mock_key_1": [42, 42],
        "mock_key_2": [42, 42],
    }


def test_json_mode_without_schema_returns_default_object(client):
    response = client.post(
        "/v1/chat/completions",
        json={
            "model": "gpt-4o",
            "response_format": {"type": "json_object"},
            "messages": [{"role": "user", "content": "Return json"}],
        },
    )

    assert response.status_code == 200
    content = response.json()["choices"][0]["message"]["content"]
    body = json.loads(content)
    assert body == {"message": "lorem ipsum"}


def test_json_invalid_scenario_breaks_json(client):
    response = client.post(
        "/v1/chat/completions",
        headers={"X-Mock-Scenario": "json_invalid"},
        json={
            "model": "gpt-4o",
            "response_format": {"type": "json_object"},
            "messages": [{"role": "user", "content": "Return json"}],
        },
    )

    assert response.status_code == 200
    content = response.json()["choices"][0]["message"]["content"]
    with pytest.raises(json.JSONDecodeError):
        json.loads(content)



from __future__ import annotations

import base64

import httpx

from myna.pytest_plugin import MynaFixture


def test_myna_fixture_helpers():
    fixture = MynaFixture(base_url="http://127.0.0.1:9999/v1", default_scenario="error=rate_limit")
    assert fixture.url("/chat/completions") == "http://127.0.0.1:9999/v1/chat/completions"
    assert fixture.headers() == {"X-Mock-Scenario": "error=rate_limit"}
    assert fixture.path_with_scenario("/audio/transcriptions").endswith(
        "scenario=error%3Drate_limit"
    )
    assert fixture.url_with_scenario("/models").endswith("scenario=error%3Drate_limit")
    assert fixture.headers("delay=500,error=server") == {
        "X-Mock-Scenario": "delay=500,error=server"
    }


def test_myna_capture_api(myna: MynaFixture):
    response = httpx.post(
        myna.url("/audio/transcriptions"),
        files={"file": ("sample.wav", b"abc", "audio/wav")},
        data={"model": "whisper-mock", "language": "nl"},
        timeout=2,
    )
    assert response.status_code == 200

    req = myna.last_request
    assert req is not None
    assert req.method == "POST"
    assert req.path == "/v1/audio/transcriptions"
    assert req.form["model"] == "whisper-mock"
    assert req.form["language"] == "nl"

    uploaded_file = req.files["file"]
    assert uploaded_file["content_type"] == "audio/wav"
    assert base64.b64decode(uploaded_file["content_base64"]) == b"abc"

    all_requests = myna.requests
    assert len(all_requests) == 1

    assert myna.clear_requests() == 1
    assert myna.last_request is None
    assert myna.clear_seeded_responses() == 0


def test_myna_url_fixture(myna: MynaFixture, myna_url: str):
    assert myna_url == myna.base_url



from __future__ import annotations

import asyncio
import random
from dataclasses import dataclass, field
from typing import Final

from fastapi.responses import JSONResponse

DEFAULT_TIMEOUT_DELAY_MS: Final[int] = 1000
VALID_ERRORS: Final[set[str]] = {"rate_limit", "auth", "context_length", "server", "timeout"}
CHAOS_ERRORS: Final[tuple[str, ...]] = ("rate_limit", "auth", "context_length", "server")


@dataclass(slots=True, frozen=True)
class MockError:
    status_code: int
    error_type: str
    message: str
    code: str | None = None


ERROR_DEFINITIONS: Final[dict[str, MockError]] = {
    "rate_limit": MockError(
        status_code=429,
        error_type="rate_limit_exceeded",
        message="Rate limit exceeded for mock request.",
        code="rate_limit_exceeded",
    ),
    "auth": MockError(
        status_code=401,
        error_type="invalid_api_key",
        message="Invalid API key provided.",
        code="invalid_api_key",
    ),
    "context_length": MockError(
        status_code=400,
        error_type="context_length_exceeded",
        message="Maximum context length exceeded.",
        code="context_length_exceeded",
    ),
    "server": MockError(
        status_code=500,
        error_type="server_error",
        message="The server had an error while processing your request.",
        code="server_error",
    ),
    "timeout": MockError(
        status_code=504,
        error_type="timeout",
        message="Upstream model request timed out.",
        code="gateway_timeout",
    ),
}


@dataclass(slots=True)
class ScenarioSpec:
    raw: str = ""
    delay_ms: int = 0
    stream_truncate: bool = False
    json_invalid: bool = False
    empty_choices: bool = False
    explicit_errors: list[str] = field(default_factory=list)
    chaos_rate: float | None = None

    @property
    def first_error(self) -> str | None:
        for error_key in self.explicit_errors:
            if error_key in VALID_ERRORS:
                return error_key
        return None


def parse_scenario(raw_scenario: str | None) -> ScenarioSpec:
    spec = ScenarioSpec(raw=(raw_scenario or "").strip())
    if not raw_scenario:
        return spec

    for directive in raw_scenario.split(","):
        token = directive.strip()
        if not token:
            continue

        if token.startswith("delay="):
            parsed_delay = _parse_non_negative_int(token.split("=", maxsplit=1)[1])
            if parsed_delay is not None:
                spec.delay_ms = parsed_delay
            continue

        if token.startswith("error="):
            error_key = token.split("=", maxsplit=1)[1].strip()
            if error_key in VALID_ERRORS:
                spec.explicit_errors.append(error_key)
            continue

        if token.startswith("chaos="):
            parsed_rate = _parse_rate(token.split("=", maxsplit=1)[1])
            if parsed_rate is not None:
                spec.chaos_rate = parsed_rate
            continue

        if token == "stream_truncate":
            spec.stream_truncate = True
            continue

        if token == "json_invalid":
            spec.json_invalid = True
            continue

        if token == "empty_choices":
            spec.empty_choices = True

    return spec


def resolve_error(spec: ScenarioSpec, global_chaos_rate: float) -> str | None:
    explicit_error = spec.first_error
    if explicit_error is not None:
        return explicit_error

    effective_chaos_rate = spec.chaos_rate if spec.chaos_rate is not None else global_chaos_rate
    if effective_chaos_rate <= 0:
        return None

    if random.random() < effective_chaos_rate:
        return random.choice(CHAOS_ERRORS)
    return None


def error_response(error_key: str) -> JSONResponse:
    error = ERROR_DEFINITIONS[error_key]
    payload = {
        "error": {
            "message": error.message,
            "type": error.error_type,
            "param": None,
            "code": error.code,
        }
    }
    return JSONResponse(status_code=error.status_code, content=payload)


async def apply_delay_ms(delay_ms: int) -> None:
    if delay_ms > 0:
        await asyncio.sleep(delay_ms / 1000)


def timeout_delay_ms(spec: ScenarioSpec) -> int:
    return spec.delay_ms if spec.delay_ms > 0 else DEFAULT_TIMEOUT_DELAY_MS


def _parse_non_negative_int(value: str) -> int | None:
    try:
        parsed = int(value)
    except ValueError:
        return None
    if parsed < 0:
        return None
    return parsed


def _parse_rate(value: str) -> float | None:
    try:
        parsed = float(value)
    except ValueError:
        return None
    if not 0 <= parsed <= 1:
        return None
    return parsed



from __future__ import annotations


def test_embeddings_respect_configured_dims(client_factory, monkeypatch):
    monkeypatch.setenv("MOCK_EMBEDDING_DIMS", "8")
    with client_factory() as client:
        response = client.post("/v1/embeddings", json={"input": ["hello", "world"]})

    assert response.status_code == 200
    payload = response.json()
    assert len(payload["data"]) == 2
    assert len(payload["data"][0]["embedding"]) == 8
    assert len(payload["data"][1]["embedding"]) == 8


def test_images_generations_support_url_and_b64(client):
    response_url = client.post("/v1/images/generations", json={"prompt": "cat", "n": 2})
    assert response_url.status_code == 200
    payload_url = response_url.json()
    assert "url" in payload_url["data"][0]
    assert len(payload_url["data"]) == 2

    response_b64 = client.post(
        "/v1/images/generations",
        json={"prompt": "cat", "response_format": "b64_json"},
    )
    assert response_b64.status_code == 200
    payload_b64 = response_b64.json()
    assert "b64_json" in payload_b64["data"][0]


def test_audio_speech_supports_wav_and_mp3(client):
    wav_response = client.post(
        "/v1/audio/speech",
        json={"model": "gpt-4o-mini-tts", "input": "hello", "response_format": "wav"},
    )
    assert wav_response.status_code == 200
    assert wav_response.headers["content-type"].startswith("audio/wav")
    assert wav_response.content.startswith(b"RIFF")

    mp3_response = client.post(
        "/v1/audio/speech",
        json={"model": "gpt-4o-mini-tts", "input": "hello"},
    )
    assert mp3_response.status_code == 200
    assert mp3_response.headers["content-type"].startswith("audio/mpeg")
    assert mp3_response.content.startswith(b"ID3")


def test_audio_transcription_and_translation(client):
    files = {"file": ("sample.wav", b"\x00\x00\x00\x00", "audio/wav")}
    transcribe_response = client.post(
        "/v1/audio/transcriptions",
        files=files,
        data={"model": "whisper-1"},
    )
    assert transcribe_response.status_code == 200
    assert transcribe_response.json()["text"] == "This is a mock transcription."

    translate_response = client.post(
        "/v1/audio/translations",
        files=files,
        data={"model": "whisper-1"},
    )
    assert translate_response.status_code == 200
    assert translate_response.json()["text"] == "This is a mock English translation."


def test_auth_toggle(client_factory, monkeypatch):
    monkeypatch.setenv("MOCK_REQUIRE_AUTH", "true")
    with client_factory() as client:
        unauthenticated = client.get("/v1/models")
        authenticated = client.get("/v1/models", headers={"Authorization": "Bearer test-token"})

    assert unauthenticated.status_code == 401
    assert unauthenticated.json()["error"]["type"] == "invalid_api_key"
    assert authenticated.status_code == 200



from __future__ import annotations


def test_models_happy_path(client):
    response = client.get("/v1/models")
    assert response.status_code == 200
    payload = response.json()
    assert payload["object"] == "list"
    assert any(model["id"] == "mock-chat-v1" for model in payload["data"])
    assert all(model["owned_by"] == "myna" for model in payload["data"])


def test_chat_completion_happy_path(client):
    response = client.post(
        "/v1/chat/completions",
        json={"model": "gpt-4o", "messages": [{"role": "user", "content": "hello world"}]},
    )

    assert response.status_code == 200
    payload = response.json()
    assert payload["object"] == "chat.completion"
    assert payload["choices"][0]["message"]["role"] == "assistant"
    assert "Mock response to: hello world" in payload["choices"][0]["message"]["content"]


def test_legacy_completion_happy_path(client):
    response = client.post(
        "/v1/completions",
        json={"model": "gpt-4o", "prompt": "write a line"},
    )

    assert response.status_code == 200
    payload = response.json()
    assert payload["object"] == "text_completion"
    assert payload["choices"][0]["finish_reason"] == "stop"
    assert "Mock response to: write a line" in payload["choices"][0]["text"]


def test_empty_choices_scenario(client):
    response = client.post(
        "/v1/chat/completions",
        headers={"X-Mock-Scenario": "empty_choices"},
        json={"model": "gpt-4o", "messages": [{"role": "user", "content": "hello"}]},
    )

    assert response.status_code == 200
    payload = response.json()
    assert payload["choices"] == []



from __future__ import annotations

from fastapi import APIRouter, Depends, File, Form, Request, UploadFile
from fastapi.responses import JSONResponse, Response

from myna.config import Settings, get_settings
from myna.generators import (
    generate_mp3_stub,
    generate_transcription_text,
    generate_translation_text,
    generate_wav_stub,
)
from myna.routers.utils import apply_scenario_or_error, parse_json_body

router = APIRouter(prefix="/v1", tags=["audio"])


@router.post("/audio/speech")
async def create_speech(
    request: Request,
    settings: Settings = Depends(get_settings),
) -> Response:
    payload = await parse_json_body(request)
    _, early_response = await apply_scenario_or_error(request, settings)
    if early_response is not None:
        return early_response

    response_format = str(payload.get("response_format") or "mp3").lower()
    if response_format == "wav":
        return Response(content=generate_wav_stub(), media_type="audio/wav")
    return Response(content=generate_mp3_stub(), media_type="audio/mpeg")


@router.post("/audio/transcriptions")
async def create_transcription(
    request: Request,
    file: UploadFile | None = File(default=None),
    model: str = Form(default="mock-asr-v1"),
    settings: Settings = Depends(get_settings),
) -> JSONResponse:
    _, early_response = await apply_scenario_or_error(request, settings)
    if early_response is not None:
        return early_response

    if file is not None:
        await file.read()
    _ = model
    return JSONResponse({"text": generate_transcription_text()})


@router.post("/audio/translations")
async def create_translation(
    request: Request,
    file: UploadFile | None = File(default=None),
    model: str = Form(default="mock-asr-v1"),
    settings: Settings = Depends(get_settings),
) -> JSONResponse:
    _, early_response = await apply_scenario_or_error(request, settings)
    if early_response is not None:
        return early_response

    if file is not None:
        await file.read()
    _ = model
    return JSONResponse({"text": generate_translation_text()})



from __future__ import annotations

from collections.abc import Callable

import pytest
from fastapi.testclient import TestClient

from myna.config import get_settings
from myna.main import create_app
from myna.response_seeding import seeded_response_queue

pytest_plugins = ["myna.pytest_plugin"]


@pytest.fixture(autouse=True)
def clear_settings_cache():
    get_settings.cache_clear()
    seeded_response_queue.clear()
    yield
    seeded_response_queue.clear()
    get_settings.cache_clear()


@pytest.fixture
def client_factory():
    def _factory() -> TestClient:
        get_settings.cache_clear()
        return TestClient(create_app())

    return _factory


@pytest.fixture
def client(client_factory: Callable[[], TestClient]):
    with client_factory() as test_client:
        yield test_client



from __future__ import annotations

import os
import socket
import subprocess
import sys
import time
from collections.abc import Iterator
from dataclasses import dataclass
from urllib.parse import quote

import httpx
import pytest


@dataclass(slots=True)
class CapturedRequest:
    method: str
    path: str
    query: dict[str, str]
    headers: dict[str, str]
    content_type: str
    body_text: str
    body_base64: str
    json: object | None
    form: dict[str, object]
    files: dict[str, object]



def _free_port() -> int:
    with socket.socket() as s:
        s.bind(("127.0.0.1", 0))
        return int(s.getsockname()[1])


@dataclass(slots=True)
class MynaFixture:
    base_url: str
    default_scenario: str | None = None

    def url(self, path: str) -> str:
        return f"{self.base_url}{self._normalize_short_path(path)}"

    def path_with_scenario(self, path: str, scenario: str | None = None) -> str:
        scenario_value = scenario if scenario is not None else self.default_scenario
        normalized_path = self._normalize_short_path(path)
        if not scenario_value:
            return normalized_path
        separator = "&" if "?" in normalized_path else "?"
        return f"{normalized_path}{separator}scenario={quote(scenario_value)}"

    def url_with_scenario(self, path: str, scenario: str | None = None) -> str:
        return f"{self.base_url}{self.path_with_scenario(path, scenario)}"

    def headers(self, scenario: str | None = None) -> dict[str, str]:
        scenario_value = scenario if scenario is not None else self.default_scenario
        if not scenario_value:
            return {}
        return {"X-Mock-Scenario": scenario_value}

    @property
    def requests(self) -> list[CapturedRequest]:
        response = httpx.get(f"{self._root_url}/__myna/requests", timeout=2)
        response.raise_for_status()
        records = response.json().get("requests", [])
        return [CapturedRequest(**record) for record in records]

    @property
    def last_request(self) -> CapturedRequest | None:
        response = httpx.get(f"{self._root_url}/__myna/requests/last", timeout=2)
        response.raise_for_status()
        payload = response.json().get("request")
        if payload is None:
            return None
        return CapturedRequest(**payload)

    def clear_requests(self) -> int:
        response = httpx.delete(f"{self._root_url}/__myna/requests", timeout=2)
        response.raise_for_status()
        return int(response.json().get("cleared", 0))

    def next_response(
        self,
        body: object,
        *,
        path: str = "/chat/completions",
        method: str = "POST",
        status_code: int = 200,
        headers: dict[str, str] | None = None,
    ) -> None:
        normalized_path = self._normalize_seed_path(path)
        response = httpx.post(
            f"{self._root_url}/__myna/responses/next",
            timeout=2,
            json={
                "path": normalized_path,
                "method": method.upper(),
                "status_code": status_code,
                "headers": headers or {},
                "json_body": body,
            },
        )
        response.raise_for_status()

    def clear_seeded_responses(self) -> int:
        response = httpx.delete(f"{self._root_url}/__myna/responses", timeout=2)
        response.raise_for_status()
        return int(response.json().get("cleared", 0))

    @property
    def _root_url(self) -> str:
        return self.base_url.removesuffix("/v1")

    @staticmethod
    def _normalize_short_path(path: str) -> str:
        return path if path.startswith("/") else f"/{path}"

    def _normalize_seed_path(self, path: str) -> str:
        normalized_path = self._normalize_short_path(path)
        if normalized_path.startswith("/v1/"):
            return normalized_path
        if normalized_path == "/v1":
            return normalized_path
        return f"/v1{normalized_path}"


@pytest.fixture(scope="session")
def myna_base_url() -> Iterator[str]:
    port = _free_port()
    env = os.environ.copy()
    pythonpath = env.get("PYTHONPATH")
    if pythonpath:
        env["PYTHONPATH"] = f"src:{pythonpath}"
    else:
        env["PYTHONPATH"] = "src"

    proc = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "uvicorn",
            "myna.main:app",
            "--host",
            "127.0.0.1",
            "--port",
            str(port),
        ],
        env=env,
    )

    health_url = f"http://127.0.0.1:{port}/healthz"
    base_url = f"http://127.0.0.1:{port}/v1"

    for _ in range(50):
        try:
            if httpx.get(health_url, timeout=0.2).status_code == 200:
                break
        except Exception:
            pass
        time.sleep(0.1)
    else:
        proc.terminate()
        raise RuntimeError("Myna fixture could not start server in time.")

    try:
        yield base_url
    finally:
        proc.terminate()
        proc.wait(timeout=5)


@pytest.fixture
def myna_scenario(request: pytest.FixtureRequest) -> str | None:
    # Supports: @pytest.mark.parametrize("myna_scenario", ["error=rate_limit"], indirect=True)
    return getattr(request, "param", None)


@pytest.fixture
def myna(myna_base_url: str, myna_scenario: str | None) -> MynaFixture:
    fixture = MynaFixture(base_url=myna_base_url, default_scenario=myna_scenario)
    fixture.clear_seeded_responses()
    fixture.clear_requests()
    return fixture


@pytest.fixture
def myna_url(myna: MynaFixture) -> str:
    return myna.base_url



from __future__ import annotations

import httpx

from myna.pytest_plugin import MynaFixture


def test_next_response_overrides_once(myna: MynaFixture):
    seeded_payload = {
        "id": "chatcmpl-seeded",
        "object": "chat.completion",
        "created": 0,
        "model": "mock-chat-v1",
        "choices": [{"index": 0, "message": {"role": "assistant", "content": ""}}],
        "usage": {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0},
    }
    myna.next_response(seeded_payload, path="/chat/completions")

    first = httpx.post(
        myna.url("/chat/completions"),
        json={"model": "mock-chat-v1", "messages": [{"role": "user", "content": "hello"}]},
        timeout=2,
    )
    assert first.status_code == 200
    assert first.json() == seeded_payload

    second = httpx.post(
        myna.url("/chat/completions"),
        json={"model": "mock-chat-v1", "messages": [{"role": "user", "content": "hello"}]},
        timeout=2,
    )
    assert second.status_code == 200
    assert second.json()["id"] != "chatcmpl-seeded"


def test_next_response_can_seed_malformed_shape(myna: MynaFixture):
    myna.next_response({"choices": [{"message": {"content": None}}]})

    response = httpx.post(
        myna.url("/chat/completions"),
        json={"model": "mock-chat-v1", "messages": [{"role": "user", "content": "hello"}]},
        timeout=2,
    )
    assert response.status_code == 200
    assert response.json()["choices"][0]["message"]["content"] is None


def test_next_response_accepts_full_path_for_compat(myna: MynaFixture):
    myna.next_response({"compat": True}, path="/v1/chat/completions")

    response = httpx.post(
        myna.url("/chat/completions"),
        json={"model": "mock-chat-v1", "messages": [{"role": "user", "content": "hello"}]},
        timeout=2,
    )
    assert response.status_code == 200
    assert response.json()["compat"] is True



from __future__ import annotations

import asyncio
import json
import time
import uuid
from typing import Any

from fastapi import APIRouter, Depends, Request
from fastapi.responses import JSONResponse, StreamingResponse

from myna.config import Settings, get_settings
from myna.generators import (
    extract_structured_json_schema,
    generate_chat_text,
    generate_json_object,
    is_json_mode,
    split_stream_chunks,
)
from myna.routers.utils import apply_scenario_or_error, parse_json_body

router = APIRouter(prefix="/v1", tags=["chat"])


@router.post("/chat/completions")
async def chat_completions(
    request: Request,
    settings: Settings = Depends(get_settings),
):
    payload = await parse_json_body(request)
    scenario, early_response = await apply_scenario_or_error(request, settings)
    if early_response is not None:
        return early_response

    model = str(payload.get("model") or settings.mock_default_model)
    completion_id = f"chatcmpl-{uuid.uuid4().hex[:12]}"
    created = int(time.time())
    content = _render_chat_content(payload, scenario.json_invalid)

    if payload.get("stream") is True:
        stream = _chat_stream(
            completion_id=completion_id,
            created=created,
            model=model,
            content=content,
            chunk_delay_seconds=settings.chunk_delay_seconds,
            stream_truncate=scenario.stream_truncate,
            empty_choices=scenario.empty_choices,
        )
        return StreamingResponse(stream, media_type="text/event-stream")

    usage = _usage(
        prompt_text=json.dumps(payload.get("messages", []), ensure_ascii=True),
        completion_text=content if not scenario.empty_choices else "",
    )
    choices: list[dict[str, Any]]
    if scenario.empty_choices:
        choices = []
    else:
        choices = [
            {
                "index": 0,
                "message": {"role": "assistant", "content": content},
                "finish_reason": "stop",
            }
        ]

    response_payload = {
        "id": completion_id,
        "object": "chat.completion",
        "created": created,
        "model": model,
        "choices": choices,
        "usage": usage,
    }
    return JSONResponse(response_payload)


@router.post("/completions")
async def completions(
    request: Request,
    settings: Settings = Depends(get_settings),
):
    payload = await parse_json_body(request)
    scenario, early_response = await apply_scenario_or_error(request, settings)
    if early_response is not None:
        return early_response

    model = str(payload.get("model") or settings.mock_default_model)
    completion_id = f"cmpl-{uuid.uuid4().hex[:12]}"
    created = int(time.time())
    text = generate_chat_text(prompt=payload.get("prompt"))

    if payload.get("stream") is True:
        stream = _completion_stream(
            completion_id=completion_id,
            created=created,
            model=model,
            text=text,
            chunk_delay_seconds=settings.chunk_delay_seconds,
            stream_truncate=scenario.stream_truncate,
            empty_choices=scenario.empty_choices,
        )
        return StreamingResponse(stream, media_type="text/event-stream")

    usage = _usage(prompt_text=str(payload.get("prompt", "")), completion_text=text)
    choices: list[dict[str, Any]]
    if scenario.empty_choices:
        choices = []
    else:
        choices = [
            {
                "index": 0,
                "text": text,
                "finish_reason": "stop",
                "logprobs": None,
            }
        ]

    response_payload = {
        "id": completion_id,
        "object": "text_completion",
        "created": created,
        "model": model,
        "choices": choices,
        "usage": usage,
    }
    return JSONResponse(response_payload)


def _render_chat_content(payload: dict[str, Any], force_invalid_json: bool) -> str:
    if is_json_mode(payload):
        if force_invalid_json:
            return '{"malformed_json": true'
        schema = extract_structured_json_schema(payload)
        generated = generate_json_object(schema)
        return json.dumps(generated, ensure_ascii=True)
    return generate_chat_text(messages=payload.get("messages"))


def _usage(prompt_text: str, completion_text: str) -> dict[str, int]:
    prompt_tokens = _rough_token_count(prompt_text)
    completion_tokens = _rough_token_count(completion_text)
    return {
        "prompt_tokens": prompt_tokens,
        "completion_tokens": completion_tokens,
        "total_tokens": prompt_tokens + completion_tokens,
    }


def _rough_token_count(text: str) -> int:
    stripped = text.strip()
    if not stripped:
        return 0
    return max(1, len(stripped.split()))


def _to_sse(data: dict[str, Any] | str) -> str:
    if isinstance(data, str):
        return f"data: {data}\n\n"
    encoded = json.dumps(data, ensure_ascii=True, separators=(",", ":"))
    return f"data: {encoded}\n\n"


async def _chat_stream(
    *,
    completion_id: str,
    created: int,
    model: str,
    content: str,
    chunk_delay_seconds: float,
    stream_truncate: bool,
    empty_choices: bool,
):
    if empty_choices:
        yield _to_sse(
            {
                "id": completion_id,
                "object": "chat.completion.chunk",
                "created": created,
                "model": model,
                "choices": [],
            }
        )
        if not stream_truncate:
            yield _to_sse("[DONE]")
        return

    yield _to_sse(
        {
            "id": completion_id,
            "object": "chat.completion.chunk",
            "created": created,
            "model": model,
            "choices": [{"index": 0, "delta": {"role": "assistant"}, "finish_reason": None}],
        }
    )

    for emitted_content_chunks, chunk in enumerate(split_stream_chunks(content), start=1):
        yield _to_sse(
            {
                "id": completion_id,
                "object": "chat.completion.chunk",
                "created": created,
                "model": model,
                "choices": [{"index": 0, "delta": {"content": chunk}, "finish_reason": None}],
            }
        )
        if stream_truncate and emitted_content_chunks >= 3:
            return
        await asyncio.sleep(chunk_delay_seconds)

    yield _to_sse(
        {
            "id": completion_id,
            "object": "chat.completion.chunk",
            "created": created,
            "model": model,
            "choices": [{"index": 0, "delta": {}, "finish_reason": "stop"}],
        }
    )
    yield _to_sse("[DONE]")


async def _completion_stream(
    *,
    completion_id: str,
    created: int,
    model: str,
    text: str,
    chunk_delay_seconds: float,
    stream_truncate: bool,
    empty_choices: bool,
):
    if empty_choices:
        yield _to_sse(
            {
                "id": completion_id,
                "object": "text_completion",
                "created": created,
                "model": model,
                "choices": [],
            }
        )
        if not stream_truncate:
            yield _to_sse("[DONE]")
        return

    for emitted_content_chunks, chunk in enumerate(split_stream_chunks(text), start=1):
        yield _to_sse(
            {
                "id": completion_id,
                "object": "text_completion",
                "created": created,
                "model": model,
                "choices": [
                    {
                        "index": 0,
                        "text": chunk,
                        "logprobs": None,
                        "finish_reason": None,
                    }
                ],
            }
        )
        if stream_truncate and emitted_content_chunks >= 3:
            return
        await asyncio.sleep(chunk_delay_seconds)

    yield _to_sse(
        {
            "id": completion_id,
            "object": "text_completion",
            "created": created,
            "model": model,
            "choices": [
                {
                    "index": 0,
                    "text": "",
                    "logprobs": None,
                    "finish_reason": "stop",
                }
            ],
        }
    )
    yield _to_sse("[DONE]")



from __future__ import annotations

import json


def test_chat_streaming_includes_done(client):
    response = client.post(
        "/v1/chat/completions",
        json={
            "model": "gpt-4o",
            "stream": True,
            "messages": [{"role": "user", "content": "hello streaming world"}],
        },
    )

    assert response.status_code == 200
    data_lines = _extract_data_lines(response.text)
    assert data_lines[-1] == "[DONE]"

    first_chunk = json.loads(data_lines[0])
    assert first_chunk["object"] == "chat.completion.chunk"
    assert first_chunk["choices"][0]["delta"]["role"] == "assistant"


def test_chat_stream_truncate_omits_done(client):
    response = client.post(
        "/v1/chat/completions",
        headers={"X-Mock-Scenario": "stream_truncate"},
        json={
            "model": "gpt-4o",
            "stream": True,
            "messages": [{"role": "user", "content": "truncate this stream please"}],
        },
    )

    assert response.status_code == 200
    data_lines = _extract_data_lines(response.text)
    assert data_lines[-1] != "[DONE]"
    assert len(data_lines) >= 4


def _extract_data_lines(stream_text: str) -> list[str]:
    result: list[str] = []
    for line in stream_text.splitlines():
        if line.startswith("data: "):
            result.append(line.removeprefix("data: "))
    return result



from __future__ import annotations

import logging

from fastapi import FastAPI, Request
from pydantic import BaseModel, ConfigDict, Field

from myna.config import get_settings
from myna.middleware.scenario_middleware import ScenarioMiddleware
from myna.request_capture import build_request_record, request_log
from myna.response_seeding import SeededResponse, seeded_response_queue
from myna.routers import audio, chat, embeddings, images, models
from myna.scenarios import error_response

logger = logging.getLogger("myna")
_INTERNAL_PREFIXES = ("/__myna/requests", "/__myna/responses")


class NextResponsePayload(BaseModel):
    model_config = ConfigDict(extra="forbid")

    path: str = Field(default="/v1/chat/completions")
    method: str = Field(default="POST")
    status_code: int = Field(default=200, ge=100, le=599)
    headers: dict[str, str] = Field(default_factory=dict)
    json_body: object | None = None
    body_text: str | None = None


def create_app() -> FastAPI:
    settings = get_settings()
    _configure_logging()

    app = FastAPI(title="myna", version="0.1.0")
    app.add_middleware(ScenarioMiddleware)

    @app.middleware("http")
    async def auth_and_request_logging(request: Request, call_next):
        is_internal_endpoint = request.url.path.startswith(_INTERNAL_PREFIXES)

        if settings.mock_log_requests:
            logger.info("%s %s", request.method, request.url.path)

        if not is_internal_endpoint:
            body = await request.body()
            request_record = await build_request_record(request, body)
            request_log.add(request_record)
            request = Request(request.scope, receive=_clone_receive(body))

        if settings.mock_require_auth and not is_internal_endpoint:
            auth_header = request.headers.get("Authorization", "")
            token = auth_header.removeprefix("Bearer ").strip()
            if not auth_header.startswith("Bearer ") or not token:
                return error_response("auth")

        if not is_internal_endpoint:
            seeded = seeded_response_queue.consume(request.method, request.url.path)
            if seeded is not None:
                return seeded.build_response()

        return await call_next(request)

    @app.get("/healthz")
    async def healthcheck() -> dict[str, str]:
        return {"status": "ok"}

    @app.get("/__myna/requests")
    async def get_requests() -> dict[str, list[dict[str, object]]]:
        return {"requests": request_log.all()}

    @app.get("/__myna/requests/last")
    async def get_last_request() -> dict[str, dict[str, object] | None]:
        return {"request": request_log.last()}

    @app.delete("/__myna/requests")
    async def clear_requests() -> dict[str, int]:
        return {"cleared": request_log.clear()}

    @app.post("/__myna/responses/next")
    async def queue_next_response(payload: NextResponsePayload) -> dict[str, bool]:
        seeded_response_queue.add(
            SeededResponse(
                method=payload.method.upper(),
                path=payload.path,
                status_code=payload.status_code,
                headers=payload.headers,
                json_body=payload.json_body,
                body_text=payload.body_text,
            )
        )
        return {"queued": True}

    @app.delete("/__myna/responses")
    async def clear_seeded_responses() -> dict[str, int]:
        return {"cleared": seeded_response_queue.clear()}

    app.include_router(models.router)
    app.include_router(chat.router)
    app.include_router(embeddings.router)
    app.include_router(images.router)
    app.include_router(audio.router)
    return app


def _clone_receive(body: bytes):
    sent = False

    async def receive() -> dict[str, object]:
        nonlocal sent
        if not sent:
            sent = True
            return {"type": "http.request", "body": body, "more_body": False}
        return {"type": "http.request", "body": b"", "more_body": False}

    return receive


def _configure_logging() -> None:
    if logger.handlers:
        return
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(name)s %(message)s")


app = create_app()



from __future__ import annotations

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import Response

from myna.scenarios import parse_scenario


class ScenarioMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next) -> Response:
        header_value = request.headers.get("X-Mock-Scenario")
        query_value = request.query_params.get("scenario")
        raw_scenario = header_value if header_value is not None else query_value
        request.state.scenario_spec = parse_scenario(raw_scenario)
        return await call_next(request)



from __future__ import annotations

from typing import Any

from fastapi import Request
from starlette.responses import Response

from myna.config import Settings
from myna.scenarios import (
    ScenarioSpec,
    apply_delay_ms,
    error_response,
    resolve_error,
    timeout_delay_ms,
)


def get_scenario(request: Request) -> ScenarioSpec:
    scenario = getattr(request.state, "scenario_spec", None)
    if isinstance(scenario, ScenarioSpec):
        return scenario
    return ScenarioSpec()


async def apply_scenario_or_error(
    request: Request,
    settings: Settings,
) -> tuple[ScenarioSpec, Response | None]:
    scenario = get_scenario(request)
    error_key = resolve_error(scenario, settings.mock_chaos_rate)
    if error_key is not None:
        if error_key == "timeout":
            await apply_delay_ms(timeout_delay_ms(scenario))
        else:
            await apply_delay_ms(scenario.delay_ms)
        return scenario, error_response(error_key)

    await apply_delay_ms(scenario.delay_ms)
    return scenario, None


async def parse_json_body(request: Request) -> dict[str, Any]:
    try:
        payload = await request.json()
    except Exception:
        return {}

    if isinstance(payload, dict):
        return payload
    return {}



from __future__ import annotations

import hashlib
import io
import random
import wave
from collections.abc import Mapping, Sequence
from typing import Any

MOCK_IMAGE_BASE64 = (
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR4nGNgYAAAAAMAASsJTYQAAAAASUVORK5CYII="
)


def generate_chat_text(
    messages: Sequence[Mapping[str, Any]] | None = None,
    prompt: Any = None,
) -> str:
    if prompt is not None:
        source = _flatten_value(prompt).strip()
    else:
        source = _last_user_message(messages).strip()

    if not source:
        return "Mock response."
    return f"Mock response to: {source[:200]}"


def split_stream_chunks(text: str) -> list[str]:
    words = text.split()
    if not words:
        return ["Mock"]

    chunks: list[str] = [words[0]]
    chunks.extend(f" {word}" for word in words[1:])
    return chunks


def is_json_mode(payload: Mapping[str, Any]) -> bool:
    response_format = payload.get("response_format")
    if not isinstance(response_format, Mapping):
        return False

    response_type = response_format.get("type")
    return response_type in {"json_object", "json_schema"}


def extract_structured_json_schema(payload: Mapping[str, Any]) -> dict[str, Any] | None:
    response_format = payload.get("response_format")
    if isinstance(response_format, Mapping):
        response_type = response_format.get("type")
        if response_type == "json_schema":
            json_schema_config = response_format.get("json_schema")
            if isinstance(json_schema_config, Mapping):
                schema = json_schema_config.get("schema")
                if isinstance(schema, Mapping):
                    return dict(schema)

        json_schema_config = response_format.get("json_schema")
        if isinstance(json_schema_config, Mapping):
            schema = json_schema_config.get("schema")
            if isinstance(schema, Mapping):
                return dict(schema)

    tools = payload.get("tools")
    if isinstance(tools, list):
        for tool in tools:
            if not isinstance(tool, Mapping):
                continue
            function_config = tool.get("function")
            if not isinstance(function_config, Mapping):
                continue
            parameters = function_config.get("parameters")
            if isinstance(parameters, Mapping):
                return dict(parameters)

    return None


def generate_json_object(schema: Mapping[str, Any] | None) -> dict[str, Any]:
    if schema is None:
        return {"message": "lorem ipsum"}

    value = _generate_from_schema(schema, depth=0)
    if isinstance(value, dict):
        return value
    return {"value": value}


def generate_embedding_vector(source: Any, dims: int) -> list[float]:
    text = _flatten_value(source)
    digest = hashlib.sha256(text.encode("utf-8")).digest()
    seed = int.from_bytes(digest[:8], byteorder="big", signed=False)
    rng = random.Random(seed)
    return [round(rng.uniform(-1.0, 1.0), 6) for _ in range(dims)]


def generate_image_stub(response_format: str, index: int) -> dict[str, str]:
    if response_format == "b64_json":
        return {"b64_json": MOCK_IMAGE_BASE64}
    return {"url": f"https://example.com/mock-image-{index + 1}.png"}


def generate_wav_stub(duration_ms: int = 300) -> bytes:
    sample_rate = 16000
    sample_count = int(sample_rate * (duration_ms / 1000))
    frames = b"\x00\x00" * sample_count

    buffer = io.BytesIO()
    with wave.open(buffer, "wb") as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(2)
        wav_file.setframerate(sample_rate)
        wav_file.writeframes(frames)
    return buffer.getvalue()


def generate_mp3_stub() -> bytes:
    # This is a tiny deterministic byte payload suitable for API contract tests.
    return b"ID3\x03\x00\x00\x00\x00\x00\x15TIT2\x00\x00\x00\x05\x00\x00Mock"


def generate_transcription_text() -> str:
    return "This is a mock transcription."


def generate_translation_text() -> str:
    return "This is a mock English translation."


def _last_user_message(messages: Sequence[Mapping[str, Any]] | None) -> str:
    if not messages:
        return ""

    for message in reversed(messages):
        if message.get("role") != "user":
            continue
        return _flatten_value(message.get("content"))
    return ""


def _flatten_value(value: Any) -> str:
    if isinstance(value, str):
        return value

    if isinstance(value, list):
        parts: list[str] = []
        for item in value:
            if isinstance(item, str):
                parts.append(item)
                continue
            if isinstance(item, Mapping):
                text = item.get("text") or item.get("input_text")
                if isinstance(text, str):
                    parts.append(text)
        return " ".join(parts).strip()

    if value is None:
        return ""
    return str(value)


def _generate_from_schema(schema: Mapping[str, Any], depth: int) -> Any:
    if depth > 5:
        return "lorem ipsum"

    enum_values = schema.get("enum")
    if isinstance(enum_values, list) and enum_values:
        return enum_values[0]

    if "default" in schema:
        return schema["default"]

    schema_type = _normalize_schema_type(schema.get("type"))

    if schema_type is None:
        for key in ("anyOf", "oneOf", "allOf"):
            variants = schema.get(key)
            if isinstance(variants, list) and variants:
                selected = _select_schema_variant(variants)
                if selected is not None:
                    return _generate_from_schema(selected, depth + 1)

        properties = schema.get("properties")
        if isinstance(properties, Mapping):
            schema_type = "object"

    if schema_type == "string":
        return _generate_string_value(schema)
    if schema_type == "integer":
        return 42
    if schema_type == "number":
        return 3.14
    if schema_type == "boolean":
        return True
    if schema_type == "array":
        items_schema = schema.get("items")
        if isinstance(items_schema, Mapping):
            return [
                _generate_from_schema(items_schema, depth + 1),
                _generate_from_schema(items_schema, depth + 1),
            ]
        return ["item_1", "item_2"]
    if schema_type == "object":
        properties = schema.get("properties")
        if not isinstance(properties, Mapping):
            additional_properties = schema.get("additionalProperties")
            if isinstance(additional_properties, Mapping):
                return {
                    "mock_key_1": _generate_from_schema(additional_properties, depth + 1),
                    "mock_key_2": _generate_from_schema(additional_properties, depth + 1),
                }
            return {"key": "lorem ipsum"}
        result: dict[str, Any] = {}
        for key, property_schema in properties.items():
            if isinstance(property_schema, Mapping):
                result[str(key)] = _generate_from_schema(property_schema, depth + 1)
            else:
                result[str(key)] = "lorem ipsum"
        return result

    return "lorem ipsum"


def _normalize_schema_type(schema_type: Any) -> str | None:
    if isinstance(schema_type, str):
        return None if schema_type == "null" else schema_type
    if isinstance(schema_type, list):
        for entry in schema_type:
            if isinstance(entry, str) and entry != "null":
                return entry
    return None


def _select_schema_variant(variants: list[Any]) -> Mapping[str, Any] | None:
    for variant in variants:
        if not isinstance(variant, Mapping):
            continue
        if _normalize_schema_type(variant.get("type")) is not None:
            return variant

    for variant in variants:
        if isinstance(variant, Mapping):
            return variant
    return None


def _generate_string_value(schema: Mapping[str, Any]) -> str:
    schema_format = schema.get("format")
    if schema_format == "date":
        return "2026-01-01"
    if schema_format == "date-time":
        return "2026-01-01T00:00:00"
    if schema_format == "email":
        return "mock@example.com"
    if schema_format in {"uri", "url"}:
        return "https://example.com"
    if schema_format == "uuid":
        return "00000000-0000-0000-0000-000000000000"
    return "lorem ipsum"



from __future__ import annotations

import time

from fastapi import APIRouter, Depends, Request
from fastapi.responses import JSONResponse

from myna.config import Settings, get_settings
from myna.generators import generate_image_stub
from myna.routers.utils import apply_scenario_or_error, parse_json_body

router = APIRouter(prefix="/v1", tags=["images"])


@router.post("/images/generations")
async def create_image(
    request: Request,
    settings: Settings = Depends(get_settings),
) -> JSONResponse:
    payload = await parse_json_body(request)
    _, early_response = await apply_scenario_or_error(request, settings)
    if early_response is not None:
        return early_response

    response_format = str(payload.get("response_format") or "url")
    requested_count = _to_positive_int(payload.get("n"), default=1)
    n = min(requested_count, 10)
    data = [generate_image_stub(response_format=response_format, index=index) for index in range(n)]
    return JSONResponse({"created": int(time.time()), "data": data})


def _to_positive_int(raw_value: object, default: int) -> int:
    try:
        parsed = int(raw_value)
    except (TypeError, ValueError):
        return default
    if parsed <= 0:
        return default
    return parsed



from __future__ import annotations

import base64
import json
from threading import Lock
from typing import Any

from starlette.datastructures import FormData, UploadFile
from starlette.requests import Request


def _request_from_body(scope: dict[str, Any], body: bytes) -> Request:
    sent = False

    async def receive() -> dict[str, Any]:
        nonlocal sent
        if not sent:
            sent = True
            return {"type": "http.request", "body": body, "more_body": False}
        return {"type": "http.request", "body": b"", "more_body": False}

    return Request(scope, receive=receive)


async def build_request_record(request: Request, body: bytes) -> dict[str, Any]:
    headers = {key: value for key, value in request.headers.items()}
    content_type = request.headers.get("content-type", "")

    record: dict[str, Any] = {
        "method": request.method,
        "path": request.url.path,
        "query": {key: value for key, value in request.query_params.items()},
        "headers": headers,
        "content_type": content_type,
        "body_text": body.decode("utf-8", errors="replace"),
        "body_base64": base64.b64encode(body).decode("ascii"),
        "json": None,
        "form": {},
        "files": {},
    }

    probe = _request_from_body(request.scope, body)

    if "application/json" in content_type:
        try:
            record["json"] = await probe.json()
        except json.JSONDecodeError:
            record["json"] = None
    elif (
        "multipart/form-data" in content_type
        or "application/x-www-form-urlencoded" in content_type
    ):
        try:
            form: FormData = await probe.form()
            _extract_form_fields(record, form)
        except Exception:
            # Keep request capture best-effort to avoid breaking test traffic.
            pass

    return record


def _extract_form_fields(record: dict[str, Any], form: FormData) -> None:
    form_fields: dict[str, Any] = {}
    files: dict[str, Any] = {}

    for key, value in form.multi_items():
        if isinstance(value, UploadFile):
            file_entry = {
                "filename": value.filename,
                "content_type": value.content_type,
            }
            content = value.file.read()
            if isinstance(content, bytes):
                file_entry["size"] = len(content)
                file_entry["content_base64"] = base64.b64encode(content).decode("ascii")
                file_entry["content_text"] = content.decode("utf-8", errors="replace")
            _append_multi_value(files, key, file_entry)
            continue

        _append_multi_value(form_fields, key, value)

    record["form"] = form_fields
    record["files"] = files


def _append_multi_value(target: dict[str, Any], key: str, value: Any) -> None:
    if key not in target:
        target[key] = value
        return

    existing = target[key]
    if isinstance(existing, list):
        existing.append(value)
        return

    target[key] = [existing, value]


class RequestLog:
    def __init__(self) -> None:
        self._requests: list[dict[str, Any]] = []
        self._lock = Lock()

    def add(self, request_record: dict[str, Any]) -> None:
        with self._lock:
            self._requests.append(request_record)

    def all(self) -> list[dict[str, Any]]:
        with self._lock:
            return list(self._requests)

    def last(self) -> dict[str, Any] | None:
        with self._lock:
            if not self._requests:
                return None
            return self._requests[-1]

    def clear(self) -> int:
        with self._lock:
            count = len(self._requests)
            self._requests.clear()
            return count


request_log = RequestLog()



from __future__ import annotations

from fastapi import APIRouter, Depends, Request
from fastapi.responses import JSONResponse

from myna.config import Settings, get_settings
from myna.generators import generate_embedding_vector
from myna.routers.utils import apply_scenario_or_error, parse_json_body

router = APIRouter(prefix="/v1", tags=["embeddings"])


@router.post("/embeddings")
async def create_embeddings(
    request: Request,
    settings: Settings = Depends(get_settings),
) -> JSONResponse:
    payload = await parse_json_body(request)
    _, early_response = await apply_scenario_or_error(request, settings)
    if early_response is not None:
        return early_response

    raw_input = payload.get("input", "")
    normalized_input = raw_input if isinstance(raw_input, list) else [raw_input]

    data = []
    for index, item in enumerate(normalized_input):
        data.append(
            {
                "object": "embedding",
                "index": index,
                "embedding": generate_embedding_vector(item, settings.mock_embedding_dims),
            }
        )

    prompt_tokens = sum(
        max(1, len(str(item).split()))
        for item in normalized_input
        if str(item).strip()
    )
    response_payload = {
        "object": "list",
        "data": data,
        "model": str(payload.get("model") or "mock-embedding-v1"),
        "usage": {"prompt_tokens": prompt_tokens, "total_tokens": prompt_tokens},
    }
    return JSONResponse(response_payload)



from __future__ import annotations


def test_request_capture_json_payload(client):
    client.delete("/__myna/requests")

    response = client.post(
        "/v1/chat/completions",
        json={"model": "gpt-4o", "messages": [{"role": "user", "content": "hello"}]},
    )
    assert response.status_code == 200

    captured = client.get("/__myna/requests/last")
    assert captured.status_code == 200
    payload = captured.json()["request"]

    assert payload["method"] == "POST"
    assert payload["path"] == "/v1/chat/completions"
    assert payload["json"]["model"] == "gpt-4o"
    assert payload["headers"]["content-type"].startswith("application/json")


def test_request_capture_multipart_payload(client):
    client.delete("/__myna/requests")

    response = client.post(
        "/v1/audio/transcriptions",
        files={"file": ("sample.wav", b"abc", "audio/wav")},
        data={"model": "whisper-1", "language": "nl"},
    )
    assert response.status_code == 200

    captured = client.get("/__myna/requests/last")
    payload = captured.json()["request"]

    assert payload["form"]["model"] == "whisper-1"
    assert payload["form"]["language"] == "nl"
    assert payload["files"]["file"]["filename"] == "sample.wav"
    assert payload["files"]["file"]["content_type"] == "audio/wav"


def test_capture_endpoints_are_not_captured(client):
    client.delete("/__myna/requests")

    response_requests = client.get("/__myna/requests")
    assert response_requests.status_code == 200

    response_seed = client.post(
        "/__myna/responses/next",
        json={"path": "/v1/chat/completions", "json_body": {"choices": []}},
    )
    assert response_seed.status_code == 200

    response_after = client.get("/__myna/requests")
    assert response_after.status_code == 200
    assert response_after.json()["requests"] == []



from __future__ import annotations

from fastapi import APIRouter, Depends, Request
from fastapi.responses import JSONResponse

from myna.config import Settings, get_settings
from myna.routers.utils import apply_scenario_or_error

router = APIRouter(prefix="/v1", tags=["models"])


@router.get("/models")
async def list_models(
    request: Request,
    settings: Settings = Depends(get_settings),
) -> JSONResponse:
    _, early_response = await apply_scenario_or_error(request, settings)
    if early_response is not None:
        return early_response

    model_ids = [
        settings.mock_default_model,
        "mock-chat-lite-v1",
        "mock-embedding-v1",
        "mock-image-v1",
        "mock-tts-v1",
        "mock-asr-v1",
    ]

    unique_model_ids = list(dict.fromkeys(model_ids))
    data = [
        {"id": model_id, "object": "model", "created": 0, "owned_by": "myna"}
        for model_id in unique_model_ids
    ]
    return JSONResponse({"object": "list", "data": data})



from __future__ import annotations

from dataclasses import dataclass, field
from threading import Lock
from typing import Any

from fastapi.responses import JSONResponse
from starlette.responses import Response


@dataclass(slots=True)
class SeededResponse:
    method: str
    path: str
    status_code: int
    headers: dict[str, str] = field(default_factory=dict)
    json_body: Any | None = None
    body_text: str | None = None

    def build_response(self) -> Response:
        if self.json_body is not None:
            return JSONResponse(
                status_code=self.status_code,
                content=self.json_body,
                headers=self.headers,
            )

        return Response(
            content=(self.body_text or ""),
            status_code=self.status_code,
            headers=self.headers,
        )


class SeededResponseQueue:
    def __init__(self) -> None:
        self._items: list[SeededResponse] = []
        self._lock = Lock()

    def add(self, seeded: SeededResponse) -> None:
        with self._lock:
            self._items.append(seeded)

    def consume(self, method: str, path: str) -> SeededResponse | None:
        match_method = method.upper()
        with self._lock:
            for index, item in enumerate(self._items):
                if item.method == match_method and item.path == path:
                    return self._items.pop(index)
        return None

    def clear(self) -> int:
        with self._lock:
            count = len(self._items)
            self._items.clear()
            return count


seeded_response_queue = SeededResponseQueue()



from __future__ import annotations

import pytest

from myna import scenarios


@pytest.mark.parametrize(
    ("scenario_value", "status_code", "error_type"),
    [
        ("error=rate_limit", 429, "rate_limit_exceeded"),
        ("error=auth", 401, "invalid_api_key"),
        ("error=context_length", 400, "context_length_exceeded"),
        ("error=server", 500, "server_error"),
        ("error=timeout,delay=1", 504, "timeout"),
    ],
)
def test_explicit_error_scenarios(client, scenario_value, status_code, error_type):
    response = client.post(
        "/v1/chat/completions",
        headers={"X-Mock-Scenario": scenario_value},
        json={"model": "gpt-4o", "messages": [{"role": "user", "content": "hi"}]},
    )

    assert response.status_code == status_code
    payload = response.json()
    assert payload["error"]["type"] == error_type


def test_header_scenario_precedence_over_query(client):
    response = client.post(
        "/v1/chat/completions?scenario=error=server",
        headers={"X-Mock-Scenario": "error=auth"},
        json={"model": "gpt-4o", "messages": [{"role": "user", "content": "hi"}]},
    )

    assert response.status_code == 401
    assert response.json()["error"]["type"] == "invalid_api_key"


def test_request_chaos_override(client, monkeypatch):
    monkeypatch.setattr(scenarios.random, "random", lambda: 0.0)
    monkeypatch.setattr(scenarios.random, "choice", lambda _: "server")

    response = client.post(
        "/v1/chat/completions",
        headers={"X-Mock-Scenario": "chaos=1"},
        json={"model": "gpt-4o", "messages": [{"role": "user", "content": "hi"}]},
    )

    assert response.status_code == 500
    assert response.json()["error"]["type"] == "server_error"


def test_global_chaos_rate(client_factory, monkeypatch):
    monkeypatch.setenv("MOCK_CHAOS_RATE", "1")
    monkeypatch.setattr(scenarios.random, "random", lambda: 0.0)
    monkeypatch.setattr(scenarios.random, "choice", lambda _: "rate_limit")

    with client_factory() as client:
        response = client.post(
            "/v1/chat/completions",
            json={"model": "gpt-4o", "messages": [{"role": "user", "content": "hi"}]},
        )

    assert response.status_code == 429
    assert response.json()["error"]["type"] == "rate_limit_exceeded"


def test_first_explicit_error_wins(client):
    response = client.post(
        "/v1/chat/completions",
        headers={"X-Mock-Scenario": "error=server,error=auth"},
        json={"model": "gpt-4o", "messages": [{"role": "user", "content": "hi"}]},
    )

    assert response.status_code == 500
    assert response.json()["error"]["type"] == "server_error"
